﻿
/*********************************************************************
 **
 **   server/appx-client-resource.js - Client Resource processing
 **
 **   This module contains code to process client resources.
 **
 *********************************************************************/

// what_str =  "@(#)Appx $Header: /src/cvs/appxHtml5/server/appx-client-resource.js,v 1.25 2015/06/12 13:16:16 pete Exp $";

//RESOURCE Message Handler
//Injects an Image into the DOM with an Object URL
//(allows server to send binary image data from the server without client saving locally
// or referencing a link to the server)

// To optimize for speed we have a number of caches we work with here.
// 1) appx_session.image_cache:
//        This is a list of the resources referenced on the current screen update.
//        This lets ApplyImages() apply generated urls to those reference elements based on class names.
// 2) AppxResource.cache[]:
//        This is an in-memory list of generated resource URLs from all resources processed during this session.
//        This lets us redisplay screens without having to reload data or regenerate URLs from blobs.
// 3) localStorage:
//        This is an on-disk persistent list of raw bytes we can load into a blob and generate a url.
//        Once loaded from here the generated url goes into the in-memory cache for the duration of the session.
//        localStorage is very limited in size so we do not store Picture Widget dynamic images as they tend to be one-off photos.
//
// FIXME: We would look at adding a 4th cache that uses the local connector to cache to disk to get past the localStorage limits.

// The flow of resources is that the screen painters "load" resources as they are encountered in the widget
// data for a screen.  The "load" function will return a unique class names that the widget elements are 
// tagged with.  The "load" function makes sure the resource is loaded and read.  It does this by looking 
// in all of it's caches and if not found sends a request to the server for the resource data.  The server
// sends these resources as Resource transactions that are processed by the "handler" function below.
// As these resources are loaded they are added to our local caches.  AppxScreen.applyImage() takes care
// of marking all screen elements that have resources with the generated URLs so they display.

function AppxResource() {}

// 
// In memory session cache for resource blob image URLs
//
AppxResource.cache = {};

//
// After the screen has been updated clear out the screen painting cache.  
// This is a temporary list of resource classes needed for this screen cycle.
//
AppxResource.done = function () {
    appx_session.image_cache = {
        "keys": [],
        "length": 0
    };
}

//
// Handler to process Resource transaction coming in from the server for resource we don't have locally yet
//
AppxResource.handler = function (x) {
    var longCacheId = appx_session.pendingResources.shift();
    var cacheid = cleanClassName(appx_session.server + "_" + appx_session.port + "_" + ab2str(x.data.ap).trim() + "_" + ab2str(x.data.ver).trim() + "_" + longCacheId.trim());
    try {
        // If we have data then process it into a resource
        if (x.data.len > 0) {
            var dataBuffer = base64ToArrayBuffer(x.data.data);

            // If we are going to need the byte array then generated it.
            if (x.data.loctype == 1 || longCacheId.length == 8) {
                var int8Array = Array.prototype.slice.call(new Int8Array(dataBuffer));
            }

            // If it's a client url then we don't have to convert from a blob. Just use the url we were given.
            if (x.data.loctype == 1) {
		if( int8Array[int8Array.length-1] == 0 )
		    int8Array[int8Array.length-1] = 32;

                var url = ab2str(int8Array).trim();
		localStorage.setItem(cacheid, url);
            }
            // Convert the blob into a url and place it in our in memory cache.
            else {
		if( x.data.state == 10 ) {
		    
		    var bb = new Int8Array(dataBuffer);
		    var bba = ab2str(bb).replace(/\r/g,"").replace(/\t/g,"").trim().replace(/^CKEDITOR[.]config[ ]*[=][ ]*/g,"").replace(/[;]$/g,"").split("\n");
		    var bbn = "";
		    for (var i = 0; i < bba.length; i++) {
			if( /^[ ]*[/][/]/g.test(bba[i]) ) {
			    // skip this line
			}
			else {
			    bbn += bba[i].replace(/[+][ ]*$/g,"").trim();
			}
		    }
		    var url = eval("("+bbn+")");
		}
		else {
		    var url = window.URL.createObjectURL(new Blob([dataBuffer]));
		}
                AppxResource.cache[cacheid] = url;

		if (longCacheId.length == 8) {
		    try {
			localStorage.setItem(cacheid, JSON.stringify(int8Array));
		    }
		    catch (ex) {
			console.log("can't save resource to localStorage: " + ex);
		    }
		}
            }
            // If it's a named resource then cache it to localStorage, we don't cache Picture Widget dynamic resources to save space.
        }
        // No data means no resource, show a broken image url
        else {
	    if( x.data.state == 10 )
		var url = "";
	    else
		var url = ""+appxClientRoot+"/images/missing.png";
        }
    }
    catch (ex) {
        console.log("image exception 1 = " + ex);
	    if( x.data.state == 10 )
		var url = "";
	    else
		var url = ""+appxClientRoot+"/images/missing.png";
    }
    
    if( x.data.state == 10 ) {
	if (appx_session.pendingResources.length == 0) {
	    appxSetStatusStateText(APPX_STATE_READY);
	    if (!screenflipped)
		appxshowscreen();
	}
    }
    else {
	// Force the image to load and apply it once loaded
	try {
	    img = new Image();
	    img.src = url;
	    appx_session.image_cache[cacheid].url = url;
	    img.onload = applyimage(cacheid);
	}
	catch (ex) {
	    console.log("appx-client-resource.js AppxResource.handler() failed to load "+cacheid+" image:" + ex);
	    applyimage(null);
	}
    }

    if( appx_session.pendingResources.length == 0 )
	AppxResource.done();
}

//
// Make sure resource is loaded and return a unique cacheid class name.
//
AppxResource.load = function (wIcon) {
    var ai = wIcon.split('.');
    var cache = appx_session.image_cache;
    var cacheid = cleanClassName(appx_session.server + "_" + appx_session.port + "_" + ai[0].replace("#", "") + "_" + ai[1] + "_" + ai[2]);
    console.log("wIcon=["+wIcon+"], cacheid=["+cacheid+"]");
    // If it's already been processed elsewhere on this screen there is nothing to do.
    if (cache[cacheid]) {
        return cacheid;
    }

    // Add it to the screens cache list
    cache.length++;
    cache.keys.push(cacheid);
    cache[cacheid] = {
        "cacheid": cacheid,
        "url": "",
        "blob": null,
        "wIcon": wIcon
    };

    // If it's not in the in memory cache try to loaded it from localStorage, if not in localStorage request it from the server.
    if (!AppxResource.cache[cacheid]) {
        if (localStorage.getItem(cacheid)) {

	    if( localStorage.getItem(cacheid).substring(0,4).toLowerCase() == "http" )
		AppxResource.cache[cacheid] = localStorage.getItem(cacheid);
	    else
		AppxResource.cache[cacheid] = window.URL.createObjectURL(new Blob([toArrayBuffer(JSON.parse(localStorage.getItem(cacheid)))]));
        }
        else {
            AppxResource.send(wIcon);
            return cacheid;
        }
    }

    // At this point we have a resource url in our in memory cache, use it.
    cache[cacheid].url = AppxResource.cache[cacheid];

    return cacheid;
}

//
// Request a resource from the server
//
AppxResource.send = function (name) {
    var cacheArray = name.split(",");
    var subArray = cacheArray[0].split(".");
    appx_session.pendingResources.push(subArray[2]);
    var ms = {
        cmd: 'appxresource',
        args: name,
        handler: 'appxresourcehandler',
        data: null
    };
    appx_session.ws.send(JSON.stringify(ms));
}
